/******************************************************************************
 * FILE: control_manager_SIL.c
 * DESCRIPTION:
 * Software in the loop simulation (SILS) environment for V&V of onboard flight program (OFP).
 *
 *
 * University of Minnesota 
 * Aerospace Engineering and Mechanics 
 * Copyright 2011 Regents of the University of Minnesota. All rights reserved.
 *
 * $Id: control_manager_SIL.c 277 2011-02-23 21:42:02Z murch $
 ******************************************************************************/

#define S_FUNCTION_NAME control_manager_SIL
#define S_FUNCTION_LEVEL 2

#include "simstruc.h"
#include "mex.h"
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <stddef.h>
#include <time.h>
#include "control_law_interface.h" // control law interfaces

#define NIN  2				/* number of input ports */
#define IP0_SIZE  17		/* size of input port 0*/
#define IP1_SIZE  2			/* size of input port 1*/
#define NOUT 1				/* number of output ports */
#define OP0_SIZE  5		    /* size of input port 0*/
#define NPAR 1				/* number of S-function parameters */
#define SW_SAMPLE_TIME        *mxGetPr(ssGetSFcnParam(S,0))		// Software sample time, as an input parameter

double references[2]; // Control references
double *commands; // Actuator commands
double feedback[17]; // Feedback signals

/*========================================================================*
 *                          Initialization                                *
 *========================================================================*/

static void mdlInitializeSizes(SimStruct *S) {
    ssSetNumSFcnParams(S, NPAR);
    if(ssGetNumSFcnParams(S) != ssGetSFcnParamsCount(S))   return;
    
    if(!ssSetNumInputPorts(S, NIN)) return;
    
    {
        int_T   i;
        for(i = 0; i < NIN; i++){
            ssSetInputPortDirectFeedThrough(S, i, 1);
            ssSetInputPortDataType(S, i, DYNAMICALLY_TYPED);
        }
        
        ssSetInputPortWidth(S, 0, IP0_SIZE);
        ssSetInputPortWidth(S, 1, IP1_SIZE);
    }
    
    if(!ssSetNumOutputPorts(S, NOUT))    return;
    {
            ssSetOutputPortWidth(S, 0, OP0_SIZE);
    }
    
    ssSetNumSampleTimes(S, 1);   
}

static void mdlInitializeSampleTimes(SimStruct *S) {
    ssSetSampleTime(S, 0, SW_SAMPLE_TIME);
    ssSetOffsetTime(S, 0, 0.0);
}

static void mdlStart(SimStruct *S) {
    reset_control(); // reset any internal states in the controller
}

/*========================================================================*
 *                              Output                                    *
 *========================================================================*/
static void mdlOutputs(SimStruct *S, int_T tid) {
    // local variables
    real_T  *y0;
        
    // outputs
    y0 =  ssGetOutputPortRealSignal(S, 0);
       
    // Assign doublet sequences to references vector
	references[0] = *ssGetInputPortRealSignalPtrs(S, 1)[0]; // roll command
    references[1] = *ssGetInputPortRealSignalPtrs(S, 1)[1]; // pitch command
    
    // Assign sensor data to feedback vector. 
    feedback[0] = *ssGetInputPortRealSignalPtrs(S, 0)[0]; // phi
    feedback[1] = *ssGetInputPortRealSignalPtrs(S, 0)[1]; // theta
    feedback[2] = *ssGetInputPortRealSignalPtrs(S, 0)[2]; // psi
    feedback[3] = *ssGetInputPortRealSignalPtrs(S, 0)[3]; // p
    feedback[4] = *ssGetInputPortRealSignalPtrs(S, 0)[4]; // q
    feedback[5] = *ssGetInputPortRealSignalPtrs(S, 0)[5]; // r
    feedback[6] = *ssGetInputPortRealSignalPtrs(S, 0)[6]; // ax
    feedback[7] = *ssGetInputPortRealSignalPtrs(S, 0)[7]; // ay
    feedback[8] = *ssGetInputPortRealSignalPtrs(S, 0)[8]; // az
	feedback[9] = *ssGetInputPortRealSignalPtrs(S, 0)[9]; // h
	feedback[10] = *ssGetInputPortRealSignalPtrs(S, 0)[10]; // ias
	feedback[11] = *ssGetInputPortRealSignalPtrs(S, 0)[11]; // lat
    feedback[12] = *ssGetInputPortRealSignalPtrs(S, 0)[12]; // lon
    feedback[13] = *ssGetInputPortRealSignalPtrs(S, 0)[13]; // alt
    feedback[14] = *ssGetInputPortRealSignalPtrs(S, 0)[14]; // vn
    feedback[15] = *ssGetInputPortRealSignalPtrs(S, 0)[15]; // ve
    feedback[16] = *ssGetInputPortRealSignalPtrs(S, 0)[16]; // vd
    
    commands = get_control_outputs(feedback, references, ssGetT(S), SW_SAMPLE_TIME);
    
    y0[0] = commands[0]; // Aileron deflection
    y0[1] = commands[1]; // Elevator deflection
    y0[2] = commands[2]; // Rudder deflection
    //y0[3] = commands[3];  // Throttle deflection *** WHEN THROTTLE IS ACTIVELY CONTROLLED, UNCOMMENT HERE ***
	y0[3] = 0.0;  // Throttle deflection zero for SIL sim (trim setting for sim is used)
	y0[4] = 0.0;  // Flap deflection, zero for now

}


/*========================================================================*
 *                              Termination                               *
 *========================================================================*/

static void mdlTerminate(SimStruct *S) {
    
}

#ifdef  MATLAB_MEX_FILE    /* Is this file being compiled as a MEX-file? */
#include "simulink.c"      /* MEX-file interface mechanism */
#else
#include "cg_sfun.h"       /* Code generation registration function */
#endif
// ========================================================================================================
