/*
 * baseline_control.c
 *
 *  Created on: Dec 27, 2010
 *      Author: Rohit Pandita, David Escobar
 * University of Minnesota 
 * Aerospace Engineering and Mechanics 
 *  2011 Regents of the University of Minnesota. All rights reserved.
 *
 * $Id: baseline_control.c 312 2011-04-01 16:15:59Z murch $
 */

#include "control_law_interface.h"
#include "control_defs.h"
#include <math.h>  //without this, SIL sim fails

/// Definition of functions: ****************************************************
static double pitch_control(double the_ref, double pitch, double pitchrate,double roll_angle, double delta_t);
static double roll_control (double phi_ref, double roll_angle, double rollrate, double delta_t);
double yaw_rate_control (double yawrate);


/// Definition of variables: ****************************************************
static double Ts = 0; // Sample time
static double integrator[2] = {0,0};
static short anti_windup[2]={1,1};

static double pitch_gain[3]  ={0.35, 0.05, 0.1};
static double roll_gain[3]={0.3,  0.05, 0.05};//10-06-07
static double REC_gain[3]={0.0,0.0,0.0};	// roll-to-elevator crossfeed gain (currently u

/// *****************************************************************************

/// ****************************************************************************************
/// Yaw rate digital controller - parameters and variables

//   y_yaw(z)      b[0] + b[1]*z^(-1) + b[2]*z^(-2)
//   --------  =  -----------------------------------
//   u_yaw(z)       1 + a[2]*z^(-1) + a[2]*z^(-2)

#define yawdamper_gain 0.03
#define k 0.7561
#define a 1.0
#define b 1.207
#define c -0.2369
static double u_yaw[3] = {0,0,0}; // input of filter { u(k), u(k-1) , u(k-2) }
static double y_yaw[3] = {0,0,0}; // output of filter { y(k), y(k-1) , y(k-2) }
static double a_yaw[3] = {1,-1*b, -1*c}; // Filter parameters
static double b_yaw[3] = {0, yawdamper_gain*k , -1*yawdamper_gain*k*a}; // Filter parameters
static double dr; // Delta rudder

//static double u4_p=0, u4_pp=0, y4_p=0, y4_pp=0;
/// ****************************************************************************************

// Angles, rates and accelarations:  ********************************
static double phi = 0; // Roll angle
static double theta = 0; // Pitch angle
static double psi = 0; // Yaw angle
static double p = 0; // Roll rate
static double q = 0; // Pitch rate
static double r = 0; // Yaw rate
static double ax = 0 ; //accel. x direction
static double ay = 0 ; //accel. y direction
static double az = 0 ; //accel. z direction
static double h = 0;
static double ias = 0;
static double lat = 0;
static double lon = 0;
static double alt = 0;
static double vn = 0;
static double ve = 0;
static double vd = 0;
/// ****************************************************************

/// References: ****************************************************
static double theta_reference;
static double phi_reference;

// *****************************************************************

/// Control outputs:  **********************************************
/// : aileron, elevator, rudder.
static double commands[4];
/// ****************************************************************

/// Return control outputs based on references and feedback signals.
extern double *get_control_outputs(double * feedback, double *reference,double time, double sample_time) {

	Ts    = sample_time;
	phi   = feedback[0];
	theta = feedback[1];
	psi   = feedback[2];
	p     = feedback[3]; // Roll rate
	q     = feedback[4]; // Pitch rate
	r     = feedback[5]; // Yaw rate
	ax    = feedback[6]; // Accel. x
	ay    = feedback[7]; // Accel. y
    az    = feedback[8]; // Accel. z
	h     = feedback[9];
	ias   = feedback[10];
	lat   = feedback[11];
	lon   = feedback[12];
	alt   = feedback[13];
	vn    = feedback[14];
	ve    = feedback[15];
	vd    = feedback[16];	

    phi_reference = reference[0];
    theta_reference = reference[1];

    commands[0] = -roll_control(phi_reference, phi, p, Ts); // Aileron deflection [rad]
    commands[1] = -pitch_control(theta_reference, theta, q, phi, Ts); // Elevator deflection [rad]
    commands[2] = yaw_rate_control(r); // Rudder deflection [rad]
    commands[3] = 0.7; // Throttle

	return commands;
}

// Pitch control law: angles in radians. Rates in rad/s. Time in seconds
static double pitch_control(double the_ref, double pitch, double pitchrate,double roll_angle, double delta_t) {
	double de;
	integrator[0] += (the_ref - pitch) * delta_t * anti_windup[0]; //pitch error integral
	de = pitch_gain[0] * (the_ref - pitch) //rad
			+ pitch_gain[2] * (-pitchrate) + pitch_gain[1] * integrator[0]
			+ REC_gain[0] * fabs(roll_angle); // Roll-Elevator crossfeed

	//maximum deflection of elevator limited to 20 degrees
	if (fabs(de) >= ELEVATOR_MAX) {
		anti_windup[0] = 0;
		de = sign(de) * ELEVATOR_MAX;
	} else
		anti_windup[0] = 1;

	return de;
}




// Roll control law: angles in radians. Rates in rad/s. Time in seconds
static double roll_control (double phi_ref, double roll_angle, double rollrate, double delta_t)
{
	// ROLL tracking controller implemented here
	double da;

	// roll attitude tracker
	integrator[1] += (phi_ref - roll_angle)*delta_t*anti_windup[1]; //roll error integral (rad)

	da  = roll_gain[0]*(phi_ref - roll_angle)          // radian
									+ roll_gain[2]*(-rollrate)
									+ roll_gain[1]*integrator[1];

	if (fabs(roll_gain[1]*integrator[1]) >= 0.2*AILERON_MAX){ anti_windup[1] = 0;}
	else anti_windup[1] = 1;

	// Saturation
	if (fabs(da) >= AILERON_MAX){
		anti_windup[1] = 0;
		da = sign(da)*AILERON_MAX;
	}
	else anti_windup[1] = 1;

	return da;
}

double yaw_rate_control (double yawrate)
{
	u_yaw[0] = yawrate;

	// Filter:
	y_yaw[0] = b_yaw[0]*u_yaw[0]+b_yaw[1]*u_yaw[1] + b_yaw[2]*u_yaw[2] - a_yaw[1]*y_yaw[1] - a_yaw[2]*y_yaw[2];

	// Updates
	u_yaw[2] = u_yaw[1];
    u_yaw[1] = u_yaw[0];
    y_yaw[2] = y_yaw[1];
    y_yaw[1] = y_yaw[0];

    //Output
    dr= y_yaw[0];

    // Saturation
	if (fabs(dr) > RUD_YAWDAMP_AUTH_MAX) {
		dr = sign(dr)*RUD_YAWDAMP_AUTH_MAX;		// yawdamper authority limited to +/-10 deg
	}
	return dr;		// rad
}

extern void reset_control(){
	integrator[0] = integrator[1] = 0; // reset integrator state to zero
	anti_windup[0] = anti_windup[1] = 1;	// reset integrator anti_windup
}

