/*
 * lqr_control.c
 *
 *  Created on: Dec 30, 2010
 *      Author: Rohit Pandita, David Escobar
 * University of Minnesota 
 * Aerospace Engineering and Mechanics 
 *  2011 Regents of the University of Minnesota. All rights reserved.
 *
 * $Id: lqr_control.c 258 2011-01-31 22:03:21Z murch $
 */

#include "control_law_interface.h"
#include "control_defs.h"
#include <math.h> //without this, SIL sim fails

/// Definition of functions *****************************************************
void roll_control_lqr (double phi_ref, double roll_angle, double p, double r, double delta_t, double *da, double *dr);
void pitch_control_lqr ( double pitch_ref, double pitch_angle, double q, double delta_t, double *de );
// ******************************************************************************

/// Definition of variables: ****************************************************
static double Ts = 0; // Sample time
static double integrator[2] = {0,0};
static short anti_windup[2]={1,1};
/// ****************************************************************

// Angles, rates and accelarations:  ********************************
static double phi = 0; // Roll angle
static double theta = 0; // Pitch angle
static double psi = 0; // Yaw angle
static double p = 0; // Roll rate
static double q = 0; // Pitch rate
static double r = 0; // Yaw rate
static double ax = 0 ; //accel. x direction
static double ay = 0 ; //accel. y direction
static double az = 0 ; //accel. z direction
static double h = 0;
static double ias = 0;
static double lat = 0;
static double lon = 0;
static double alt = 0;
static double vn = 0;
static double ve = 0;
static double vd = 0;
/// ****************************************************************

/// Control outputs:  **********************************************
// : aileron, elevator, rudder.
static double commands[4];
static double delta_elev=0, delta_ail=0, delta_rudd=0 , delta_thr =0;
//  ****************************************************************

/// References: ****************************************************
static double theta_reference;
static double phi_reference;
// *****************************************************************

/// Return control outputs based on references and feedback signals.
extern double *get_control_outputs(double *feedback, double *reference, double time, double sample_time) {

	Ts    = sample_time;
	phi   = feedback[0];
	theta = feedback[1];
	psi   = feedback[2];
	p     = feedback[3]; // Roll rate
	q     = feedback[4]; // Pitch rate
	r     = feedback[5]; // Yaw rate
	ax    = feedback[6]; // Accel. x
	ay    = feedback[7]; // Accel. y
    az    = feedback[8]; // Accel. z
	h     = feedback[9];
	ias   = feedback[10];
	lat   = feedback[11];
	lon   = feedback[12];
	alt   = feedback[13];
	vn    = feedback[14];
	ve    = feedback[15];
	vd    = feedback[16];	

    phi_reference = reference[0];
    theta_reference = reference[1];

    pitch_control_lqr(theta_reference, theta, q, Ts, &delta_elev); // Elevator deflection [rad]
    roll_control_lqr(phi_reference, phi, p , r , Ts , &delta_ail, &delta_rudd); // Aileron and rudder deflections [rad]
    delta_thr = 0.7;

    commands[0] =  delta_ail;
    commands[1] =  delta_elev;
    commands[2] =  delta_rudd;
    commands[3] =  delta_thr; // Throttle

	return commands;
}


void roll_control_lqr (double phi_ref, double roll_angle, double p, double r, double delta_t, double *da, double *dr)
{
	// ROLL tracking controller implemented here
	double da_tmp, dr_tmp;

	double K_lqr[2][4] = {{-0.6679,   -0.0243,   -0.0251,   -0.3631},
			{-0.1355,   -0.0004,   -0.0284,   -0.0710}};		// low bandwidth

	integrator[1] += (roll_angle - phi_ref)*delta_t*anti_windup[1]; //roll error integral (rad)

	// u = -Kx
	da_tmp  = -(K_lqr[0][0]*integrator[1] + K_lqr[0][1]*p + K_lqr[0][2]*r + K_lqr[0][3]*roll_angle);
	dr_tmp  = -(K_lqr[1][0]*integrator[1] + K_lqr[1][1]*p + K_lqr[1][2]*r + K_lqr[1][3]*roll_angle);

	if (fabs(integrator[1]) >= 0.2*AILERON_MAX){ anti_windup[1] = 0;}
	else anti_windup[1] = 1;

	if (fabs(da_tmp) >= AILERON_MAX){
		anti_windup[1] = 0;
		da_tmp = sign(da_tmp)*AILERON_MAX;
	}
	else if (fabs(dr_tmp) >= RUDDER_MAX){
		anti_windup[1] = 0;
		dr_tmp = sign(dr_tmp)*RUDDER_MAX;
	}
	else anti_windup[1] = 1;

	*da = da_tmp;
	*dr = dr_tmp;
}


void pitch_control_lqr ( double pitch_ref, double pitch_angle, double q, double delta_t, double *de )
{

	     // PITCH tracking controller implemented here
	     double de_tmp;
	     // double K_lqr[3] = {-0.6592, -0.4298, -0.8051};
	     double K_lqr[3] = {-0.240144526553189,  -0.079132432193614, -0.287138908852059};	// from rohit 10/6/10

	     integrator[0] += (pitch_angle - pitch_ref)*delta_t*anti_windup[0]; //pitch error integral (rad)

	     // u = -Kx
	     de_tmp  = -(K_lqr[0]*integrator[0] + K_lqr[1]*q + K_lqr[2]*pitch_angle);

	     if (fabs(integrator[0]) >= 0.2*ELEVATOR_MAX){ anti_windup[0] = 0;}
	     else anti_windup[0] = 1;

	     // Saturation
	     if (fabs(de_tmp) >= ELEVATOR_MAX){
	         anti_windup[0] = 0;
	         de_tmp = sign(de_tmp)*ELEVATOR_MAX;
	     }
	     else anti_windup[0] = 1;

	     *de = de_tmp;
	     //*de = pitch_ref;
}

extern void reset_control(){
	integrator[0] = integrator[1] = 0; // reset integrator state to zero
	anti_windup[0] = anti_windup[1] = 1;	// reset integrator anti_windup
}
