/*
 * FILE: sensors.c
 * DESCRIPTION:
 *
 *
 *  Created on: Feb 27, 2010
 *      Author: Zhefeng Li, Curt Olson, Rohit Pandita
 *
 * University of Minnesota 
 * Aerospace Engineering and Mechanics 
 *  2011 Regents of the University of Minnesota. All rights reserved.
 *
 * $Id: sensors.c 253 2011-01-20 22:01:35Z murch $
 */

#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <time.h>
#include <cyg/io/serialio.h>
#include <cyg/io/io.h>
#include <cyg/kernel/kapi.h>
#include <cyg/io/spipsc.h>
#include <cyg/io/i2c_mpc5xxx.h>

#include "sensors.h"


// Initialize SPI communication
int init_spi() {
	cyg_uint32 len;
	cyg_uint32 cr,sicr,clken,ccr,ctlr,ctur,rfalarm,tfalarm,cdm_mclk;
	byte *p8;
	cyg_uint16 *p16;
	cyg_uint32 *p32;

	p8=MPC5XXX_PSC1+MPC5XXX_PSC_CR;
	cr=*p8;

	p32=MPC5XXX_PSC1+MPC5XXX_PSC_SICR;
	sicr=*p32;

	p32=MPC5XXX_CDM+MPC5XXX_CDM_CLKEN;
	clken=*p32;

	p32=MPC5XXX_PSC1+MPC5XXX_PSC_CCR;
	ccr=*p32;

	p8=MPC5XXX_PSC1+MPC5XXX_PSC_CTLR;
	ctlr=*p8;

	p8=MPC5XXX_PSC1+MPC5XXX_PSC_CTUR;
	ctur=*p8;

	p16=MPC5XXX_PSC1+MPC5XXX_PSC_RFALARM;
	rfalarm=*p16;

	p16=MPC5XXX_PSC1+MPC5XXX_PSC_TFALARM;
	tfalarm=*p16;

	p32=MPC5XXX_CDM+MPC5XXX_CDM_PSC1_MCLK_EN;
	cdm_mclk=*p32;

#ifdef verbose
	fprintf( stderr, "Initializing SPI.\n" );
	fprintf( stderr, " CR:%02X \n", cr);
	fprintf( stderr, " SICR:%08X \n", sicr);
	fprintf( stderr, " CLKEN:%08X \n", clken);
	fprintf( stderr, " CCR:%08X \n", ccr);
	fprintf( stderr, " CTLR:%02X \n", ctlr);
	fprintf( stderr, " CTUR:%02X \n", ctur);
	fprintf( stderr, " RFALARM:%04X \n", rfalarm);
	fprintf( stderr, " TFALARM:%04X \n", tfalarm);
	fprintf( stderr, " CDM_MCLK_EN:%08X \n", cdm_mclk);

	p8=MPC5XXX_PSC3+MPC5XXX_PSC_CR;
	cr=*p8;
	fprintf( stderr, " PSC3 CR:%02X \n", cr );
#endif

	/*
	 *  The test below opens the SPI(PSC) device, then reads some
	 *  basic configuration information from the device.
	 *
	 *  This information include buffer length, buffer counter,
	 *  non-block information,etc.
	 *
	 *  Then this program will write some bytes into the out buffer,
	 *  then check how many bytes received. Comparing the send out
	 *  bytes and received bytes to make sure the SPI's Transceiver
	 *  works well.
	 *
	 *  The next step to test is to check whether the device driver
	 *  works well with different register lengths.
	 */

	cyg_spipsc_info_t config;
	Cyg_ErrNo ret;

	ret = cyg_io_lookup( SPIPSC_PORT1, &hSPI );

	if ( ENOERR != ret ) {
#ifdef verbose
		fprintf( stderr, "open spipsc: unable to open %s - %s\n",
				SPIPSC_PORT1, strerror(ret) );
#endif
		return(-1);
	}

	len = sizeof(config);
	//Setting the baud rate
	cyg_io_get_config( hSPI, CYG_IO_GET_CONFIG_SPIPSC_INFO, &config, &len );

#ifdef verbose
	fprintf( stderr, "Open the SPI port successfully!\n" );
	fprintf( stderr, "CONFIG information dump:\n" );
	fprintf( stderr, " sclk_rate:%d\n", config.sclk_rate );
	fprintf( stderr, " master:%d\n", config.master );
	fprintf( stderr, " sclk_pol:%d\n", config.sclk_pol );
	fprintf( stderr, " sclk_phase:%d\n", config.sclk_phase );
	fprintf( stderr, " msb_first:%d\n", config.msb_first );
	fprintf( stderr, " reg_len:%d\n", config.reg_len );
	fprintf( stderr, " rxalarm_level:%d\n", config.rxalarm_level );
	fprintf( stderr, " txalarm_level:%d\n", config.txalarm_level );
	fprintf( stderr, " cs_dw_udly:%d\n", config.cs_dw_udly );
	fprintf( stderr, " tr_bt_udly:%d\n", config.tr_bt_udly );

	fprintf( stderr, "\nStarting to test the Buffer Blocking setting\n" );
#endif

	cyg_uint32 read_blk_status,write_blk_status;
	len = sizeof( cyg_uint32 );

	cyg_io_get_config( hSPI, CYG_IO_GET_CONFIG_READ_BLOCKING,
			&read_blk_status, &len);
#ifdef verbose
	fprintf( stderr, "Initial READ BLOCKING status:%d\n", read_blk_status );
#endif
	cyg_io_get_config( hSPI, CYG_IO_GET_CONFIG_WRITE_BLOCKING,
			&write_blk_status, &len );
#ifdef verbose
	fprintf( stderr, "Initial WRITE BLOCKING status:%d\n", write_blk_status );
	fprintf( stderr, "\nChange the setting\n" );
#endif

	if ( write_blk_status == 1 )
		write_blk_status = 0;
	else
		write_blk_status = 1;

	if ( read_blk_status == 1 )
		read_blk_status = 0;
	else
		read_blk_status = 1;

	//read_blk_status = 1;	// set blocking mode
	//read_blk_status = 0;	// set non-blocking mode

	cyg_io_set_config( hSPI, CYG_IO_SET_CONFIG_READ_BLOCKING,
			&read_blk_status, &len );
	cyg_io_set_config( hSPI, CYG_IO_SET_CONFIG_WRITE_BLOCKING,
			&write_blk_status, &len );
#ifdef verbose
	fprintf( stderr, "After Change the BLOCK setting:\n", read_blk_status );
#endif
	cyg_io_get_config( hSPI, CYG_IO_GET_CONFIG_READ_BLOCKING,
			&read_blk_status, &len );
#ifdef verbose
	fprintf( stderr, "READ BLOCKING status:%d\n", read_blk_status );
#endif
	cyg_io_get_config( hSPI, CYG_IO_GET_CONFIG_WRITE_BLOCKING,
			&write_blk_status, &len);
#ifdef verbose
	fprintf( stderr, "WRITE BLOCKING status:%d\n", write_blk_status );
#endif
	cyg_spipsc_buf_info_t buf_info;
	len = sizeof(buf_info);

	cyg_io_get_config( hSPI, CYG_IO_GET_CONFIG_SPIPSC_BUFFER_INFO,
			&buf_info, &len );
#ifdef verbose
	fprintf( stderr, "\nBUFFER INFO:\n" );
	fprintf( stderr, " rx_bufsize:%d\n", buf_info.rx_bufsize );
	fprintf( stderr, " rx_count:%d\n", buf_info.rx_count );
	fprintf( stderr, " tx_bufsize:%d\n", buf_info.tx_bufsize );
	fprintf( stderr, " tx_count:%d\n", buf_info.tx_count );
#endif

	return 0;
}

double convert_ad( cyg_uint16 raw, double factor ) {
	int tmp = raw;
	if ( raw > 0x1FFF ) {
		tmp = -(0x4000 - raw);
	}
	return (double)tmp * factor;
}

int read_spi(struct imu *imupkt) {
	int i;
	byte input[500]={0,};
	static cyg_uint16 rcv_reg[100]={0,};
	cyg_uint32 len, result;
	cyg_spipsc_buf_info_t buf_info;

	/*
	 * For Real sensor test
	 * Individual register data output
	 */
	cyg_uint16 data[100];

	for ( i = 0; i < 12; i++ ) {
		data[i] = 0x0200 * (i+1);
	}
	data[13] = 0x0200;
	len = 12*2;

	/*
	 * For Real sensor test
	 * Burst register data output
	 */
	cyg_io_write( hSPI, data, &len);
#ifdef verbose
	fprintf( stderr, "read_spi() Waiting......\n" );
#endif
	cyg_thread_delay(5);

	cyg_io_get_config( hSPI, CYG_IO_GET_CONFIG_SPIPSC_BUFFER_INFO,
			&buf_info, &len );


	result = len = buf_info.rx_count;
#ifdef verbose
	fprintf( stderr, "Will READ back %d bytes from the RCV buffer.\n", len );
#endif
	cyg_io_read( hSPI, input, &result );

	if ( len != result ) {
		imupkt->err_type = got_invalid;
#ifdef verbose
		fprintf( stderr, "Error: read back %d bytes from the rcv buffer, expected %d.\n", \
				result, len );
#endif
		return(-1);
	}
	if ( len != 24 ) {
		imupkt->err_type = got_invalid;
#ifdef verbose
		fprintf( stderr,"Error: didn't read 24 SPI bytes back, instead only %d were available\n", \
				result );
#endif
		return(-1);
	}
	else if ( len == 24 ) {
		imupkt->err_type = data_valid;
#ifdef verbose
		fprintf( stderr,"Success: Read 24 SPI bytes back" );
#endif
	}

	if ( result > 1 ) {
		for ( i = 0; i < result; i += 2 ) {
			rcv_reg[i/2] = (input[i] & 0x3F) * 256 + input[i+1];
		}
	}

	// update imupacket
	// *IMP* IMU axis alignment is different: Z=-Z_body, Y=-Y_body
	imupkt->Vs = convert_ad(rcv_reg[0], 0.002418);	//unit: Volt
	imupkt->p  = D2R * convert_ad(rcv_reg[2], 0.05);		//unit: rad/s
	imupkt->q  = -D2R * convert_ad(rcv_reg[3], 0.05);
	imupkt->r  = -D2R * convert_ad(rcv_reg[4], 0.05);
	imupkt->ax = -g * convert_ad(rcv_reg[5], 0.00333);		//unit: m/s^2
	imupkt->ay = g * convert_ad(rcv_reg[6], 0.00333);
	imupkt->az = g * convert_ad(rcv_reg[7], 0.00333);
	imupkt->hx = convert_ad(rcv_reg[8], 0.0005);			//unit: Gauss
	imupkt->hy = convert_ad(rcv_reg[9], 0.0005);
	imupkt->hz = convert_ad(rcv_reg[10], 0.0005);
	imupkt->T  = 25.0 + convert_ad(rcv_reg[11], 0.14);
	imupkt->time = get_Time();		//timestamp

	// output real units
#ifdef verbose
		fprintf( stderr, "\n" );
		fprintf( stderr, "power = %.2f V\n", imupkt->Vs );
		fprintf( stderr, "X gyro = %.2f deg/s\n", imupkt->p );
		fprintf( stderr, "Y gyro = %.2f deg/s\n", imupkt->q );
		fprintf( stderr, "Z gyro = %.2f deg/s\n", imupkt->r );
		fprintf( stderr, "X accel = %.2f G\n", imupkt->ax );
		fprintf( stderr, "Y accel = %.2f G\n", imupkt->ay );
		fprintf( stderr, "Z accel = %.2f G\n", imupkt->az );
		fprintf( stderr, "X mag = %.2f mgauss\n", imupkt->hx );
		fprintf( stderr, "Y mag = %.2f mgauss\n", imupkt->hy );
		fprintf( stderr, "Z mag = %.2f mgauss\n", imupkt->hz );
		fprintf( stderr, "Temp = %.1f C\n",	imupkt->T );
#endif

	return 0;
}

int	i2c_read(cyg_i2c_device* device, unsigned short start_addr,unsigned char *rx_data, unsigned int count)
{
	unsigned char	Buf[2];
	unsigned int	ret_count;

	Buf[0] = (start_addr>>8) & 0xFF;
	Buf[1] = start_addr & 0xFF;

	cyg_i2c_transaction_begin(device);
	ret_count = cyg_i2c_transaction_rx(device, true, rx_data, count, true, true);
	//then the sequential read (count-1)
	cyg_i2c_transaction_end(device);

	return ret_count;
}

void init_i2c()
{
	myAltimeter = new_HoneywellAltimeter( &i2c_honeywell_static, 0, ABS_PRESS_MAX );
	myPitot = new_HoneywellPitot( &i2c_honeywell_pitot, -DIFF_PRESS_MAX, DIFF_PRESS_MAX );
}

void read_pspd (struct imu *imupkt)
{
	int ret_count_pitot, ret_count_static;

	ret_count_pitot = HoneywellPitot_readPressure( myPitot );
	cyg_thread_delay(2);
	ret_count_static = HoneywellAltimeter_readPressure( myAltimeter );

/*
	double currentPs = HoneywellAltimeter_getPressure(myAltimeter);
	double hi = HoneywellAltimeter_getAltitude(myAltimeter);
	fprintf(stderr,"\n Bytes read = %d, current Ps= %f, current height = %f"\
			,ret_count_static, currentPs, hi);

	double currentPd = HoneywellPitot_getPressure(myPitot);
	double currentSpeed = HoneywellPitot_getSpeed(myPitot);
	fprintf(stderr,"\n Bytes read = %d, current Pd= %f, current speed = %f"\
	,ret_count_pitot, currentPd, currentSpeed);
*/
	imupkt->Ps = HoneywellAltimeter_getPressure( myAltimeter );
	imupkt->Pd = HoneywellPitot_getPressure( myPitot );
	
	imupkt->h = HoneywellAltimeter_getAltitude( myAltimeter );
	imupkt->ias = HoneywellPitot_getSpeed( myPitot );
}

/// Read microbotics air data sysmtem variables
int readAirDatyaSys(struct imu *imupkt){
  
  return 0;
}
